"""Utilities."""

import re
from typing import Optional, Tuple

NAMESPACES = {
    "didl_lite": "urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/",
    "dc": "http://purl.org/dc/elements/1.1/",
    "sec": "http://www.sec.co.kr/",
    "upnp": "urn:schemas-upnp-org:metadata-1-0/upnp/",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
}


def expand_namespace_tag(tag: str) -> str:
    """
    Expand namespace-alias to url.

    E.g.,
        _ns_tag('didl_lite:item') -> '{urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/}item'
    """
    if ":" not in tag:
        return tag

    namespace, tag = tag.split(":")
    namespace_uri = NAMESPACES[namespace]
    return f"{{{namespace_uri}}}{tag}"


def split_namespace_tag(namespaced_tag: str) -> Tuple[Optional[str], str]:
    """
    Extract namespace and tag from namespaced-tag.

    E.g., _namespace_tag('{urn:schemas-upnp-org:metadata-1-0/upnp/}class') ->
        'urn:schemas-upnp-org:metadata-1-0/upnp/', 'class'
    """
    if "}" not in namespaced_tag:
        return None, namespaced_tag

    idx = namespaced_tag.index("}")
    namespace = namespaced_tag[1:idx]
    tag = namespaced_tag[idx + 1 :]  # noqa: E203
    return namespace, tag


def to_camel_case(name: str) -> str:
    """Get camel case of name."""
    sub1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", sub1).lower()


def didl_property_key(didl_property_name: str) -> str:
    """Get Python property key for a DIDL property name."""
    if ":" in didl_property_name:
        # Remove the namespace from the property name
        didl_property_name = didl_property_name.partition(":")[2]

    if didl_property_name.startswith("@"):
        return to_camel_case(didl_property_name.replace("@", ""))

    return to_camel_case(didl_property_name.replace("@", "_"))


def didl_property_def_key(didl_property_def: Tuple[str, ...]) -> str:
    """Get Python property key for didl_property_def."""
    if didl_property_def[1].startswith("@"):
        return to_camel_case(didl_property_def[1].replace("@", ""))

    return to_camel_case(didl_property_def[1].replace("@", "_"))



"""Unit tests for didl_lite."""

import pytest
from defusedxml import ElementTree as ET

from didl_lite import didl_lite

NAMESPACES = {
    "didl_lite": "urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/",
    "dc": "http://purl.org/dc/elements/1.1/",
    "upnp": "urn:schemas-upnp-org:metadata-1-0/upnp/",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
}


class TestDidlLite:
    """Tests for didl_lite."""

    # pylint: disable=too-many-public-methods

    def test_item_from_xml(self) -> None:
        """Test item from XML."""
        didl_string = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:sec="http://www.sec.co.kr/">
    <item id="0" parentID="0" restricted="1">
        <dc:title>Audio Item Title</dc:title>
        <upnp:class>object.item.audioItem</upnp:class>
        <dc:language>English</dc:language>
        <res protocolInfo="protocol_info">url</res>
        <upnp:longDescription>Long description</upnp:longDescription>
    </item>
</DIDL-Lite>"""
        items = didl_lite.from_xml_string(didl_string)
        assert len(items) == 1

        item = items[0]
        assert item.xml_el is not None
        assert getattr(item, "title") == "Audio Item Title"
        assert getattr(item, "upnp_class") == "object.item.audioItem"
        assert getattr(item, "language") == "English"
        assert getattr(item, "longDescription") == "Long description"
        assert getattr(item, "long_description") == "Long description"
        assert isinstance(item, didl_lite.AudioItem)
        assert not hasattr(item, "non_existing")

        resources = item.res
        assert len(resources) == 1
        resource = resources[0]
        assert resource.xml_el is not None
        assert resource.protocol_info == "protocol_info"
        assert resource.uri == "url"
        assert not hasattr(item, "non_existing")
        assert item.res == item.resources

    def test_item_bad_class(self) -> None:
        """Test item from XML that has a badly-cased upnp:class."""
        didl_string = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:sec="http://www.sec.co.kr/">
    <item id="0" parentID="0" restricted="1">
        <dc:title>Audio Item Title</dc:title>
        <upnp:class>Object.Item.AudioItem</upnp:class>
        <dc:language>English</dc:language>
        <res protocolInfo="protocol_info">url</res>
        <upnp:longDescription>Long description</upnp:longDescription>
    </item>
</DIDL-Lite>"""
        with pytest.raises(
            didl_lite.DidlLiteException,
            match="upnp:class Object.Item.AudioItem is unknown",
        ):
            didl_lite.from_xml_string(didl_string)

        items = didl_lite.from_xml_string(didl_string, strict=False)
        assert len(items) == 1

        item = items[0]
        assert isinstance(item, didl_lite.AudioItem)

    def test_item_from_xml_not_strict(self) -> None:
        """Test item from XML."""
        didl_string = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:sec="http://www.sec.co.kr/">
    <item id="0" parentID="0" restricted="1">
        <dc:title>Audio Item Title</dc:title>
        <upnp:class>object.item.audioItem</upnp:class>
        <dc:language>English</dc:language>
        <res>url</res>
    </item>
</DIDL-Lite>"""
        items = didl_lite.from_xml_string(didl_string, strict=False)
        assert len(items) == 1

        item = items[0]
        assert item.xml_el is not None
        assert getattr(item, "title") == "Audio Item Title"
        assert getattr(item, "upnp_class") == "object.item.audioItem"
        assert getattr(item, "language") == "English"
        assert isinstance(item, didl_lite.AudioItem)
        assert not hasattr(item, "non_existing")

        resources = item.res
        assert len(resources) == 1
        resource = resources[0]
        assert resource.xml_el is not None
        assert resource.protocol_info is None  # This is now allowed with strict=False
        assert resource.uri == "url"
        assert not hasattr(item, "non_existing")
        assert item.res == item.resources

    def test_item_to_xml(self) -> None:
        """Test item to XML."""
        resource = didl_lite.Resource("url", "protocol_info")
        items = [
            didl_lite.AudioItem(
                id="0",
                parent_id="0",
                title="Audio Item Title",
                restricted="1",
                resources=[resource],
                language="English",
                longDescription="Long description",
            ),
        ]
        didl_string = didl_lite.to_xml_string(*items).decode("utf-8")

        assert 'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"' in didl_string
        assert 'xmlns:dc="http://purl.org/dc/elements/1.1/"' in didl_string
        assert 'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"' in didl_string
        assert 'xmlns:sec="http://www.sec.co.kr/"' in didl_string
        assert 'xmlns:ns1="urn:schemas-upnp-org:metadata-1-0/upnp/"' not in didl_string

        didl_el = ET.fromstring(didl_string)

        item_el = didl_el.find("./didl_lite:item", NAMESPACES)
        assert item_el is not None
        assert item_el.attrib["id"] == "0"
        assert item_el.attrib["parentID"] == "0"
        assert item_el.attrib["restricted"] == "1"

        title_el = item_el.find("./dc:title", NAMESPACES)
        assert title_el is not None
        assert title_el.text == "Audio Item Title"

        class_el = item_el.find("./upnp:class", NAMESPACES)
        assert class_el is not None
        assert class_el.text == "object.item.audioItem"

        language_el = item_el.find("./dc:language", NAMESPACES)
        assert language_el is not None
        assert language_el.text == "English"

        long_description_el = item_el.find("./upnp:longDescription", NAMESPACES)
        assert long_description_el is not None
        assert long_description_el.text == "Long description"

        res_el = item_el.find("./didl_lite:res", NAMESPACES)
        assert res_el is not None
        assert res_el.attrib["protocolInfo"] == "protocol_info"
        assert res_el.text == "url"

    def test_item_repr(self) -> None:
        """Test item's repr can convert back to an equivalent item."""
        # pylint: disable=import-outside-toplevel
        # repr method doesn't know how package was imported, so only uses class names
        from didl_lite.didl_lite import AudioItem, Resource

        item = AudioItem(
            id="0",
            parent_id="0",
            title="Audio Item Title",
            restricted="1",
            res=[
                Resource("url", "protocol_info"),
                Resource("url2", "protocol_info2"),
            ],
            language="English",
        )

        item_repr = repr(item)
        item_remade = eval(item_repr)  # pylint: disable=eval-used
        assert ET.tostring(item.to_xml()) == ET.tostring(item_remade.to_xml())

    def test_container_from_xml(self) -> None:
        """Test container from XML."""
        didl_string = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:sec="http://www.sec.co.kr/">
    <container id="0" parentID="0" restricted="1">
        <dc:title>Album Container Title</dc:title>
        <upnp:class>object.container.album</upnp:class>

        <item id="1" parentID="0" restricted="1">
            <dc:title>Audio Item Title</dc:title>
            <upnp:class>object.item.audioItem</upnp:class>
            <dc:language>English</dc:language>
            <res protocolInfo="protocol_info">url</res>
        </item>
    </container>
</DIDL-Lite>"""
        items = didl_lite.from_xml_string(didl_string)
        assert len(items) == 1

        container = items[0]
        assert container.xml_el is not None
        assert isinstance(container, didl_lite.Container)
        assert getattr(container, "title") == "Album Container Title"
        assert getattr(container, "upnp_class") == "object.container.album"

        item = container[0]
        assert item.xml_el is not None
        assert item.title == "Audio Item Title"
        assert item.upnp_class == "object.item.audioItem"
        assert item.language == "English"

        resources = item.res
        assert len(resources) == 1
        resource = resources[0]
        assert resource.xml_el is not None
        assert resource.protocol_info == "protocol_info"
        assert resource.uri == "url"
        assert item.res == item.resources

    def test_container_to_xml(self) -> None:
        """Test container to XML."""
        container = didl_lite.Album(id="0", parent_id="0", title="Audio Item Title", restricted="1")
        resource = didl_lite.Resource("url", "protocol_info")
        item = didl_lite.AudioItem(
            id="0",
            parent_id="0",
            title="Audio Item Title",
            restricted="1",
            resources=[resource],
            language="English",
        )
        container.append(item)

        didl_string = didl_lite.to_xml_string(container).decode("utf-8")
        didl_el = ET.fromstring(didl_string)

        container_el = didl_el.find("./didl_lite:container", NAMESPACES)
        assert container_el is not None
        assert container_el.attrib["id"] == "0"
        assert container_el.attrib["parentID"] == "0"
        assert container_el.attrib["restricted"] == "1"

        item_el = container_el.find("./didl_lite:item", NAMESPACES)
        assert item_el is not None
        assert item_el.attrib["id"] == "0"
        assert item_el.attrib["parentID"] == "0"
        assert item_el.attrib["restricted"] == "1"

        title_el = item_el.find("./dc:title", NAMESPACES)
        assert title_el is not None
        assert title_el.text == "Audio Item Title"

        class_el = item_el.find("./upnp:class", NAMESPACES)
        assert class_el is not None
        assert class_el.text == "object.item.audioItem"

        language_el = item_el.find("./dc:language", NAMESPACES)
        assert language_el is not None
        assert language_el.text == "English"

        res_el = item_el.find("./didl_lite:res", NAMESPACES)
        assert res_el is not None
        assert res_el.attrib["protocolInfo"] == "protocol_info"
        assert res_el.text == "url"

    def test_container_repr(self) -> None:
        """Test containers's repr can convert back to an equivalent container."""
        # pylint: disable=import-outside-toplevel
        from didl_lite.didl_lite import Album, AudioItem, Resource

        container = Album(id="0", parent_id="0", title="Audio Item Title", restricted="1")
        resource = Resource("url", "protocol_info")
        item = AudioItem(
            id="0",
            parent_id="0",
            title="Audio Item Title",
            restricted="1",
            resources=[resource],
            language="English",
        )
        container.append(item)

        container_repr = repr(container)
        container_remade = eval(container_repr)  # pylint: disable=eval-used
        assert ET.tostring(container.to_xml()) == ET.tostring(container_remade.to_xml())

    def test_descriptor_from_xml_root(self) -> None:
        """Test root descriptor from XML."""
        didl_string = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:sec="http://www.sec.co.kr/">
    <desc id="1" nameSpace="ns" type="type">Text</desc>
</DIDL-Lite>"""

        items = didl_lite.from_xml_string(didl_string)
        assert len(items) == 1

        descriptor = items[0]
        assert descriptor is not None
        assert descriptor.xml_el is not None
        assert getattr(descriptor, "id") == "1"
        assert getattr(descriptor, "name_space") == "ns"
        assert getattr(descriptor, "type") == "type"
        assert getattr(descriptor, "text") == "Text"

    def test_descriptor_from_xml_item(self) -> None:
        """Test item descriptor from XML."""
        didl_string = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:sec="http://www.sec.co.kr/">
    <item id="1" parentID="0" restricted="1">
        <dc:title>Audio Item Title</dc:title>
        <upnp:class>object.item.audioItem</upnp:class>
        <dc:language>English</dc:language>
        <res protocolInfo="protocol_info">url</res>
        <desc id="1" nameSpace="ns" type="type">Text</desc>
    </item>
</DIDL-Lite>"""

        items = didl_lite.from_xml_string(didl_string)
        assert len(items) == 1

        item = items[0]
        assert item is not None
        assert isinstance(item, didl_lite.AudioItem)

        descriptor = item.descriptors[0]
        assert descriptor is not None
        assert descriptor.xml_el is not None
        assert descriptor.id == "1"
        assert descriptor.name_space == "ns"
        assert descriptor.type == "type"
        assert descriptor.text == "Text"

    def test_descriptor_from_xml_container(self) -> None:
        """Test container descriptor from XML."""
        didl_string = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:sec="http://www.sec.co.kr/">
    <container id="0" parentID="0" restricted="1">
        <dc:title>Album Container Title</dc:title>
        <upnp:class>object.container.album</upnp:class>

        <desc id="1" nameSpace="ns" type="type">Text</desc>
    </container>
</DIDL-Lite>"""

        items = didl_lite.from_xml_string(didl_string)
        assert len(items) == 1

        container = items[0]
        assert container is not None
        assert container.xml_el is not None
        assert isinstance(container, didl_lite.Container)

        descriptor = container.descriptors[0]
        assert descriptor is not None
        assert descriptor.xml_el is not None
        assert descriptor.id == "1"
        assert descriptor.name_space == "ns"
        assert descriptor.type == "type"
        assert descriptor.text == "Text"

    def test_descriptor_from_xml_container_item(self) -> None:
        """Test item descriptor in container from XML."""
        didl_string = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:sec="http://www.sec.co.kr/">
    <container id="0" parentID="0" restricted="1">
        <dc:title>Album Container Title</dc:title>
        <upnp:class>object.container.album</upnp:class>

        <item id="1" parentID="0" restricted="1">
            <dc:title>Audio Item Title</dc:title>
            <upnp:class>object.item.audioItem</upnp:class>
            <dc:language>English</dc:language>
            <res protocolInfo="protocol_info">url</res>
            <desc id="1" nameSpace="ns" type="type">Text</desc>
        </item>
    </container>
</DIDL-Lite>"""

        items = didl_lite.from_xml_string(didl_string)
        assert len(items) == 1

        container = items[0]
        assert container is not None
        assert isinstance(container, didl_lite.Container)

        item = container[0]
        assert item is not None

        descriptor = item.descriptors[0]
        assert descriptor is not None
        assert descriptor.xml_el is not None
        assert descriptor.id == "1"
        assert descriptor.name_space == "ns"
        assert descriptor.type == "type"
        assert descriptor.text == "Text"

    def test_descriptor_to_xml(self) -> None:
        """Test descriptor to XML."""
        descriptor = didl_lite.Descriptor(id="1", name_space="ns", type="type", text="Text")
        item = didl_lite.AudioItem(
            id="0",
            parent_id="0",
            title="Audio Item Title",
            restricted="1",
            language="English",
            descriptors=[descriptor],
        )
        didl_string = didl_lite.to_xml_string(item).decode("utf-8")
        didl_el = ET.fromstring(didl_string)

        item_el = didl_el.find("./didl_lite:item", NAMESPACES)
        assert item_el is not None

        descriptor_el = item_el.find("./didl_lite:desc", NAMESPACES)
        assert descriptor_el is not None
        assert len(descriptor_el.attrib) == 3
        assert descriptor_el.attrib["id"] == "1"
        assert descriptor_el.attrib["nameSpace"] == "ns"
        assert descriptor_el.attrib["type"] == "type"
        assert descriptor_el.text == "Text"

        descriptor = didl_lite.Descriptor(id="2", name_space="ns2")
        descriptor_el = descriptor.to_xml()
        assert descriptor_el is not None
        assert len(descriptor_el.attrib) == 2
        assert descriptor_el.attrib["id"] == "2"
        assert descriptor_el.attrib["nameSpace"] == "ns2"

    def test_descriptor_repr(self) -> None:
        """Test descriptor's repr can convert back to an equivalent descriptorb."""
        # pylint: disable=import-outside-toplevel
        from didl_lite.didl_lite import Descriptor

        descriptor = Descriptor(id="1", name_space="ns", type="type", text="Text")

        descriptor_repr = repr(descriptor)
        descriptor_remade = eval(descriptor_repr)  # pylint: disable=eval-used
        assert ET.tostring(descriptor.to_xml()) == ET.tostring(descriptor_remade.to_xml())

    def test_item_order(self) -> None:
        """Test item ordering."""
        didl_string = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:sec="http://www.sec.co.kr/">
    <item id="0" parentID="0" restricted="1">
        <dc:title>Audio Item Title 1</dc:title>
        <upnp:class>object.item.audioItem</upnp:class>
        <dc:language>English</dc:language>
    </item>
    <container id="1" parentID="0" restricted="1">
        <dc:title>Album Container Title</dc:title>
        <upnp:class>object.container.album</upnp:class>
    </container>
    <item id="2" parentID="0" restricted="1">
        <dc:title>Audio Item Title 1</dc:title>
        <upnp:class>object.item.audioItem</upnp:class>
        <dc:language>English</dc:language>
    </item>
</DIDL-Lite>"""

        items = didl_lite.from_xml_string(didl_string)
        assert len(items) == 3

        assert isinstance(items[0], didl_lite.AudioItem)
        assert isinstance(items[1], didl_lite.Album)
        assert isinstance(items[2], didl_lite.AudioItem)

    def test_item_property_attribute_from_xml(self) -> None:
        """Test item property from XML attribute."""
        didl_string = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:sec="http://www.sec.co.kr/">
    <item id="0" parentID="0" restricted="1">
        <dc:title>Video Item Title</dc:title>
        <upnp:class>object.item.videoItem</upnp:class>
        <upnp:genre id="genreId">Action</upnp:genre>
    </item>
</DIDL-Lite>"""

        items = didl_lite.from_xml_string(didl_string)
        assert len(items) == 1

        item = items[0]
        assert item is not None
        assert getattr(item, "genre") == "Action"
        assert getattr(item, "genre_id") == "genreId"

    def test_item_property_attribute_to_xml(self) -> None:
        """Test item property to XML."""
        item = didl_lite.VideoItem(
            id="0",
            parent_id="0",
            title="Video Item Title",
            restricted="1",
            genre="Action",
            genre_id="genreId",
        )
        didl_string = didl_lite.to_xml_string(item).decode("utf-8")
        didl_el = ET.fromstring(didl_string)

        item_el = didl_el.find("./didl_lite:item", NAMESPACES)
        assert item_el is not None

        genre_el = item_el.find("./upnp:genre", NAMESPACES)
        assert genre_el is not None
        assert genre_el.text == "Action"
        assert genre_el.attrib["id"] == "genreId"

    def test_item_missing_id(self) -> None:
        """Test item missing ID from XML."""
        didl_string = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:sec="http://www.sec.co.kr/">
    <item restricted="1">
        <dc:title>Video Item Title</dc:title>
        <upnp:class>object.item</upnp:class>
    </item>
</DIDL-Lite>"""

        items = didl_lite.from_xml_string(didl_string)
        assert len(items) == 1

    def test_item_set_attributes(self) -> None:
        """Test item attribute from XML."""
        didl_string = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
           xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
           xmlns:dc="http://purl.org/dc/elements/1.1/"
           xmlns:sec="http://www.sec.co.kr/">
    <item restricted="1">
        <dc:title>Video Item Title</dc:title>
        <upnp:class>object.item.videoItem</upnp:class>
    </item>
</DIDL-Lite>"""

        items = didl_lite.from_xml_string(didl_string)
        assert len(items) == 1

        item = items[0]
        assert getattr(item, "title") == "Video Item Title"
        assert hasattr(item, "rating")
        assert getattr(item, "rating") is None
        assert isinstance(item, didl_lite.VideoItem)
        assert len(item.res) == 0
        assert item.res == item.resources

    def test_extra_properties(self) -> None:
        """Test extra item properties from XML."""
        didl_string = """
        <DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
                   xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
                   xmlns:dc="http://purl.org/dc/elements/1.1/"
                   xmlns:sec="http://www.sec.co.kr/">
            <item restricted="1">
                <dc:title>Video Item Title</dc:title>
                <upnp:class>object.item.videoItem</upnp:class>
                <upnp:albumArtURI>extra_property</upnp:albumArtURI>
            </item>
        </DIDL-Lite>"""
        items = didl_lite.from_xml_string(didl_string)
        assert len(items) == 1

        item = items[0]
        assert hasattr(item, "album_art_uri")
        assert getattr(item, "album_art_uri") == "extra_property"
        assert getattr(item, "albumArtURI") == "extra_property"

    def test_default_properties_set(self) -> None:
        """Test defaults for item properties."""
        item = didl_lite.VideoItem(id="0", parent_id="0", title="Video Item Title", restricted="1")
        assert hasattr(item, "genre_type")  # property is set

    def test_property_case(self) -> None:
        """Test item properties can be accessed using snake_case or camelCase."""
        item = didl_lite.MusicTrack(
            id="0",
            parent_id="0",
            title="Audio Item Title",
            restricted="1",
            language="English",
            originalTrackNumber="1",
            storage_medium="HDD",
        )
        assert hasattr(item, "original_track_number")
        assert hasattr(item, "originalTrackNumber")
        assert item.original_track_number is item.originalTrackNumber
        assert item.original_track_number == "1"

        assert hasattr(item, "storage_medium")
        assert hasattr(item, "storageMedium")
        assert item.storage_medium is item.storageMedium
        assert item.storage_medium == "HDD"

        assert hasattr(item, "long_description")
        assert hasattr(item, "longDescription")
        assert item.long_description is item.longDescription
        assert item.long_description is None

        assert not hasattr(item, "otherItem")
        assert not hasattr(item, "other_item")

        item.storageMedium = "CD"
        assert item.storage_medium is item.storageMedium
        assert item.storage_medium == "CD"

        item.long_description = "Long description"
        assert item.long_description is item.longDescription
        assert item.long_description == "Long description"

        item.otherItem = "otherItem"
        assert hasattr(item, "otherItem")
        assert not hasattr(item, "other_item")
        assert item.otherItem == "otherItem"

    def test_item_improper_class_nesting(self) -> None:
        """
        Test item from XML that has upnp_class element above item.

        Cater for WiiM Pro and possibly other Linkplay devices that
        emit upnp_class above the item element instead of inside it
        """
        didl_string = """
<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
    xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">
    <upnp:class>object.item.audioItem.musicTrack</upnp:class>
    <item id="0">
        <res protocolInfo="protocol_info"></res>
        <dc:title>Music Track Title</dc:title>
        <dc:creator>Music Track Creator</dc:creator>
        <upnp:artist>Artist</upnp:artist>
        <upnp:album>Album</upnp:album>
    </item>
</DIDL-Lite>"""
        items = didl_lite.from_xml_string(didl_string, strict=False)
        assert len(items) == 1

        item = items[0]
        assert isinstance(item, didl_lite.MusicTrack)

    def test_from_xml_string_unbound_prefix(self) -> None:
        """Test from_xml_string with unbound namespace prefix."""
        # The key fix is adding: xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
        broken_xml = (
            '<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" '
            'xmlns:dc="http://purl.org/dc/elements/1.1/" '
            'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
            'xmlns:dlna="urn:schemas-dlna-org:metadata-1-0/">'
            '<item id="1" parentID="0" restricted="1">'
            "<dc:title>Test Title</dc:title>"
            "<song:subTitle>Test Subtitle</song:subTitle>"
            "<upnp:class>object.item.audioItem.musicTrack</upnp:class>"
            "</item>"
            "</DIDL-Lite>"
        )

        # This call should not throw ParseError when strict=False.
        objs = didl_lite.from_xml_string(broken_xml, strict=False)

        assert len(objs) == 1
        assert objs[0].title == "Test Title"
        # Check that the temporary namespace is correctly assigned.
        assert "sub_title" in objs[0].__dict__
        assert objs[0].sub_title == "Test Subtitle"
        assert isinstance(objs[0], didl_lite.MusicTrack)

    def test_from_xml_string_unbound_prefix_without_dlna_namespace(self) -> None:
        """Test unbound prefix recovery when xmlns:dlna is absent.

        Regression: the previous implementation anchored the namespace
        injection on an existing `xmlns:dlna` declaration. Devices such as
        JBL Authentics and WiiM/LinkPlay players emit `<song:*>` tags
        without declaring `xmlns:dlna`, leaving the unbound prefix in place
        and breaking parsing even with strict=False.
        """
        broken_xml = (
            '<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" '
            'xmlns:dc="http://purl.org/dc/elements/1.1/" '
            'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">'
            '<item id="1" parentID="0" restricted="1">'
            "<dc:title>Test Title</dc:title>"
            "<song:subTitle>Test Subtitle</song:subTitle>"
            "<upnp:class>object.item.audioItem.musicTrack</upnp:class>"
            "</item>"
            "</DIDL-Lite>"
        )

        objs = didl_lite.from_xml_string(broken_xml, strict=False)

        assert len(objs) == 1
        assert objs[0].title == "Test Title"
        assert "sub_title" in objs[0].__dict__
        assert objs[0].sub_title == "Test Subtitle"
        assert isinstance(objs[0], didl_lite.MusicTrack)

    def test_from_xml_string_unbound_prefix_with_prefixed_root(self) -> None:
        """Test unbound prefix recovery when the root element carries a prefix.

        The DIDL-Lite root may be written as `<didl:DIDL-Lite>` rather than
        `<DIDL-Lite>`. The namespace injection has to recognise that form and
        write the matched name back unchanged, otherwise the opening tag no
        longer matches its closing tag.
        """
        broken_xml = (
            '<didl:DIDL-Lite xmlns:didl="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" '
            'xmlns:dc="http://purl.org/dc/elements/1.1/" '
            'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">'
            '<didl:item id="1" parentID="0" restricted="1">'
            "<dc:title>Test Title</dc:title>"
            "<song:subTitle>Test Subtitle</song:subTitle>"
            "<upnp:class>object.item.audioItem.musicTrack</upnp:class>"
            "</didl:item>"
            "</didl:DIDL-Lite>"
        )

        objs = didl_lite.from_xml_string(broken_xml, strict=False)

        assert len(objs) == 1
        assert objs[0].title == "Test Title"
        assert objs[0].sub_title == "Test Subtitle"
        assert isinstance(objs[0], didl_lite.MusicTrack)

    def test_from_xml_string_unbound_prefix_with_ncname_characters(self) -> None:
        """Test unbound prefix recovery for prefixes holding '-', '.' or '_'.

        An XML namespace prefix is an NCName, so it may contain a hyphen, a
        period or an underscore after its first character. Recovery has to
        detect those prefixes as well, not only alphanumeric ones.
        """
        for prefix in ("my-song", "my.song", "my_song"):
            broken_xml = (
                '<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" '
                'xmlns:dc="http://purl.org/dc/elements/1.1/" '
                'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">'
                '<item id="1" parentID="0" restricted="1">'
                "<dc:title>Test Title</dc:title>"
                f"<{prefix}:subTitle>Test Subtitle</{prefix}:subTitle>"
                "<upnp:class>object.item.audioItem.musicTrack</upnp:class>"
                "</item>"
                "</DIDL-Lite>"
            )

            objs = didl_lite.from_xml_string(broken_xml, strict=False)

            assert len(objs) == 1, f"prefix {prefix} was not recovered"
            assert objs[0].title == "Test Title"
            assert isinstance(objs[0], didl_lite.MusicTrack)

    def test_music_track_artist_and_genre(self) -> None:
        """Test MusicTrack artist and genre properties."""
        track = didl_lite.MusicTrack(
            id="1",
            parent_id="0",
            title="Test",
            restricted="0",
            artist="My Artist",
            genre="My Genre",
        )

        xml = didl_lite.to_xml_string(track)

        assert b"<upnp:artist>My Artist</upnp:artist>" in xml
        assert b"<upnp:genre>My Genre</upnp:genre>" in xml



"""DIDL-Lite (Digital Item Declaration Language) tools for Python."""
# pylint: disable=too-many-lines

import re
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from xml.etree import ElementTree as ET

import defusedxml.ElementTree

from .utils import (
    NAMESPACES,
    didl_property_def_key,
    didl_property_key,
    expand_namespace_tag,
    split_namespace_tag,
    to_camel_case,
)

TDO = TypeVar("TDO", bound="DidlObject")  # pylint: disable=invalid-name
TC = TypeVar("TC", bound="Container")  # pylint: disable=invalid-name
TD = TypeVar("TD", bound="Descriptor")  # pylint: disable=invalid-name
TR = TypeVar("TR", bound="Resource")  # pylint: disable=invalid-name


class DidlLiteException(Exception):
    """DIDL Lite Exception."""


# region: DidlObjects

# upnp_class to python type mapping
_upnp_class_map: Dict[str, Type["DidlObject"]] = {}
_upnp_class_map_lowercase: Dict[str, Type["DidlObject"]] = {}


class DidlObject:
    """DIDL Object."""

    tag: Optional[str] = None
    upnp_class: str = "object"
    didl_properties_defs: List[Tuple[str, str, str]] = [
        ("didl_lite", "@id", "R"),
        ("didl_lite", "@parentID", "R"),
        ("didl_lite", "@restricted", "R"),
        ("dc", "title", "R"),
        ("upnp", "class", "R"),
        ("dc", "creator", "O"),
        ("didl_lite", "res", "O"),
        ("upnp", "writeStatus", "O"),
    ]

    id: str
    parent_id: str
    res: List["Resource"]
    xml_el: Optional[ET.Element]
    descriptors: Sequence["Descriptor"]

    @classmethod
    def __init_subclass__(cls: Type["DidlObject"], **kwargs: Any) -> None:
        """Create mapping of upnp_class to Python type for fast lookup."""
        super().__init_subclass__(**kwargs)
        assert cls.upnp_class not in _upnp_class_map
        assert cls.upnp_class.lower() not in _upnp_class_map_lowercase
        _upnp_class_map[cls.upnp_class] = cls
        _upnp_class_map_lowercase[cls.upnp_class.lower()] = cls

    def __init__(
        self,
        id: str = "",
        parent_id: str = "",
        descriptors: Optional[Sequence["Descriptor"]] = None,
        xml_el: Optional[ET.Element] = None,
        strict: bool = True,
        **properties: Any,
    ) -> None:
        """Initialize."""
        # pylint: disable=invalid-name,redefined-builtin,too-many-arguments,too-many-positional-arguments
        properties["id"] = id
        properties["parent_id"] = parent_id
        properties["class"] = self.upnp_class
        properties["res"] = properties.get("res") or properties.get("resources") or []
        if "resources" in properties:
            del properties["resources"]
        self._ensure_required_properties(strict, properties)
        self._set_property_defaults()
        self._set_properties(properties)

        self.xml_el = xml_el
        self.descriptors = descriptors if descriptors else []

    def _ensure_required_properties(self, strict: bool, properties: Mapping[str, Any]) -> None:
        """Check if all required properties are given."""
        if not strict:
            return

        python_property_keys = {didl_property_key(key) for key in properties}

        for property_def in self.didl_properties_defs:
            key = didl_property_def_key(property_def)
            if property_def[2] == "R" and key not in python_property_keys:
                raise DidlLiteException(key + " is mandatory")

    def _set_property_defaults(self) -> None:
        """Ensure we have default/known slots, and set them all to None."""
        for property_def in self.didl_properties_defs:
            key = didl_property_def_key(property_def)
            setattr(self, key, None)

    def _set_properties(self, properties: Mapping[str, Any]) -> None:
        """Set attributes from properties."""
        for key, value in properties.items():
            setattr(self, key, value)

    @classmethod
    def from_xml(cls: Type[TDO], xml_el: ET.Element, strict: bool = True) -> TDO:
        """
        Initialize from an XML node.

        I.e., parse XML and return instance.
        """
        # pylint: disable=too-many-locals
        properties = {}  # type: Dict[str, Any]

        # attributes
        for attr_key, attr_value in xml_el.attrib.items():
            key = to_camel_case(attr_key)
            properties[key] = attr_value

        # child-nodes
        for xml_child_node in xml_el:
            if xml_child_node.tag == expand_namespace_tag("didl_lite:res"):
                continue

            _, tag = split_namespace_tag(xml_child_node.tag)
            key = to_camel_case(tag)
            value = xml_child_node.text
            properties[key] = value

            # attributes of child nodes
            parent_key = key
            for attr_key, attr_value in xml_child_node.attrib.items():
                key = parent_key + "_" + to_camel_case(attr_key)
                properties[key] = attr_value

        # resources
        resources = []
        for res_el in xml_el.findall("./didl_lite:res", NAMESPACES):
            resource = Resource.from_xml(res_el)
            resources.append(resource)
        properties["res"] = properties["resources"] = resources

        # descriptors
        descriptors = []
        for desc_el in xml_el.findall("./didl_lite:desc", NAMESPACES):
            descriptor = Descriptor.from_xml(desc_el)
            descriptors.append(descriptor)

        return cls(xml_el=xml_el, descriptors=descriptors, strict=strict, **properties)

    def to_xml(self) -> ET.Element:
        """Convert self to XML Element."""
        assert self.tag is not None
        item_el = ET.Element(self.tag)
        elements = {"": item_el}

        # properties
        for property_def in self.didl_properties_defs:
            if "@" in property_def[1]:
                continue
            key = didl_property_def_key(property_def)

            if getattr(self, key) is None or key == "res":  # no resources, handled later on
                continue

            tag = property_def[0] + ":" + property_def[1]
            property_el = ET.Element(tag, {})
            property_el.text = getattr(self, key)
            item_el.append(property_el)
            elements[property_def[1]] = property_el

        # attributes and property@attributes
        for property_def in self.didl_properties_defs:
            if "@" not in property_def[1]:
                continue

            key = didl_property_def_key(property_def)
            value = getattr(self, key)
            if value is None:
                continue

            el_name, attr_name = property_def[1].split("@")
            property_el = elements[el_name]
            property_el.attrib[attr_name] = value

        # resource
        for resource in self.resources:
            res_el = resource.to_xml()
            item_el.append(res_el)

        # descriptor
        for descriptor in self.descriptors:
            desc_el = descriptor.to_xml()
            item_el.append(desc_el)

        return item_el

    def __getattr__(self, name: str) -> Any:
        """Get attribute, modifying case as needed."""
        if name == "resources":
            return getattr(self, "res")
        if name in self.__dict__:
            return self.__dict__[name]
        cleaned_name = didl_property_key(name)
        if cleaned_name not in self.__dict__:
            raise AttributeError(name)
        return self.__dict__[cleaned_name]

    def __setattr__(self, name: str, value: Any) -> None:
        """Set attribute, modifying case as needed."""
        if name not in self.__dict__:
            # Redirect to the lower_camel_case version if it's already set,
            # which is the case for all defined didl properties.
            cleaned_name = didl_property_key(name)
            if cleaned_name in self.__dict__:
                name = cleaned_name
        self.__dict__[name] = value

    def __repr__(self) -> str:
        """Evaluatable string representation of this object."""
        class_name = type(self).__name__
        attr = ", ".join(f"{key}={val!r}" for key, val in self.__dict__.items() if key not in ("class", "xml_el"))
        return f"{class_name}({attr})"


# region: items
class Item(DidlObject):
    """DIDL Item."""

    # pylint: disable=too-few-public-methods

    tag = "item"
    upnp_class = "object.item"
    didl_properties_defs = DidlObject.didl_properties_defs + [
        ("didl_lite", "@refID", "O"),  # actually, R, but ignore for now
        ("upnp", "bookmarkID", "O"),
    ]


class ImageItem(Item):
    """DIDL Image Item."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.imageItem"
    didl_properties_defs = Item.didl_properties_defs + [
        ("upnp", "longDescription", "O"),
        ("upnp", "storageMedium", "O"),
        ("upnp", "rating", "O"),
        ("dc", "description", "O"),
        ("dc", "publisher", "O"),
        ("dc", "date", "O"),
        ("dc", "rights", "O"),
    ]


class Photo(ImageItem):
    """DIDL Photo."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.imageItem.photo"
    didl_properties_defs = ImageItem.didl_properties_defs + [
        ("upnp", "album", "O"),
    ]


class AudioItem(Item):
    """DIDL Audio Item."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.audioItem"
    didl_properties_defs = Item.didl_properties_defs + [
        ("upnp", "genre", "O"),
        ("dc", "description", "O"),
        ("upnp", "longDescription", "O"),
        ("dc", "publisher", "O"),
        ("dc", "language", "O"),
        ("dc", "relation", "O"),
        ("dc", "rights", "O"),
    ]


class MusicTrack(AudioItem):
    """DIDL Music Track."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.audioItem.musicTrack"
    didl_properties_defs = AudioItem.didl_properties_defs + [
        ("upnp", "artist", "O"),
        ("upnp", "album", "O"),
        ("upnp", "albumArtURI", "O"),
        ("upnp", "originalTrackNumber", "O"),
        ("upnp", "playlist", "O"),
        ("upnp", "storageMedium", "O"),
        ("dc", "contributor", "O"),
        ("dc", "date", "O"),
    ]


class AudioBroadcast(AudioItem):
    """DIDL Audio Broadcast."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.audioItem.audioBroadcast"
    didl_properties_defs = AudioItem.didl_properties_defs + [
        ("upnp", "region", "O"),
        ("upnp", "radioCallSign", "O"),
        ("upnp", "radioStationID", "O"),
        ("upnp", "radioBand", "O"),
        ("upnp", "channelNr", "O"),
        ("upnp", "signalStrength", "O"),
        ("upnp", "signalLocked", "O"),
        ("upnp", "tuned", "O"),
        ("upnp", "recordable", "O"),
    ]


class AudioBook(AudioItem):
    """DIDL Audio Book."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.audioItem.audioBook"
    didl_properties_defs = AudioItem.didl_properties_defs + [
        ("upnp", "storageMedium", "O"),
        ("upnp", "producer", "O"),
        ("dc", "contributor", "O"),
        ("dc", "date", "O"),
    ]


class VideoItem(Item):
    """DIDL Video Item."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.videoItem"
    didl_properties_defs = Item.didl_properties_defs + [
        ("upnp", "genre", "O"),
        ("upnp", "genre@id", "O"),
        ("upnp", "genre@type", "O"),
        ("upnp", "longDescription", "O"),
        ("upnp", "producer", "O"),
        ("upnp", "rating", "O"),
        ("upnp", "actor", "O"),
        ("upnp", "director", "O"),
        ("dc", "description", "O"),
        ("dc", "publisher", "O"),
        ("dc", "language", "O"),
        ("dc", "relation", "O"),
        ("upnp", "playbackCount", "O"),
        ("upnp", "lastPlaybackTime", "O"),
        ("upnp", "lastPlaybackPosition", "O"),
        ("upnp", "recordedDayOfWeek", "O"),
        ("upnp", "srsRecordScheduleID", "O"),
    ]


class Movie(VideoItem):
    """DIDL Movie."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.videoItem.movie"
    didl_properties_defs = VideoItem.didl_properties_defs + [
        ("upnp", "storageMedium", "O"),
        ("upnp", "DVDRegionCode", "O"),
        ("upnp", "channelName", "O"),
        ("upnp", "scheduledStartTime", "O"),
        ("upnp", "scheduledEndTime", "O"),
        ("upnp", "programTitle", "O"),
        ("upnp", "seriesTitle", "O"),
        ("upnp", "episodeCount", "O"),
        ("upnp", "episodeNumber", "O"),
        ("upnp", "episodeSeason", "O"),
    ]


class VideoBroadcast(VideoItem):
    """DIDL Video Broadcast."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.videoItem.videoBroadcast"
    didl_properties_defs = VideoItem.didl_properties_defs + [
        ("upnp", "icon", "O"),
        ("upnp", "region", "O"),
        ("upnp", "channelNr", "O"),
        ("upnp", "signalStrength", "O"),
        ("upnp", "signalLocked", "O"),
        ("upnp", "tuned", "O"),
        ("upnp", "recordable", "O"),
        ("upnp", "callSign", "O"),
        ("upnp", "price", "O"),
        ("upnp", "payPerView", "O"),
    ]


class MusicVideoClip(VideoItem):
    """DIDL Music Video Clip."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.videoItem.musicVideoClip"
    didl_properties_defs = VideoItem.didl_properties_defs + [
        ("upnp", "artist", "O"),
        ("upnp", "storageMedium", "O"),
        ("upnp", "album", "O"),
        ("upnp", "scheduledStartTime", "O"),
        ("upnp", "scheduledStopTime", "O"),
        # ('upnp', 'director', 'O'),  # duplicate in standard
        ("dc", "contributor", "O"),
        ("dc", "date", "O"),
    ]


class PlaylistItem(Item):
    """DIDL Playlist Item."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.playlistItem"
    didl_properties_defs = Item.didl_properties_defs + [
        ("upnp", "artist", "O"),
        ("upnp", "genre", "O"),
        ("upnp", "longDescription", "O"),
        ("upnp", "storageMedium", "O"),
        ("dc", "description", "O"),
        ("dc", "date", "O"),
        ("dc", "language", "O"),
    ]


class TextItem(Item):
    """DIDL Text Item."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.textItem"
    didl_properties_defs = Item.didl_properties_defs + [
        ("upnp", "author", "O"),
        ("upnp", "res@protection", "O"),
        ("upnp", "longDescription", "O"),
        ("upnp", "storageMedium", "O"),
        ("upnp", "rating", "O"),
        ("dc", "description", "O"),
        ("dc", "publisher", "O"),
        ("dc", "contributor", "O"),
        ("dc", "date", "O"),
        ("dc", "relation", "O"),
        ("dc", "language", "O"),
        ("dc", "rights", "O"),
    ]


class BookmarkItem(Item):
    """DIDL Bookmark Item."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.bookmarkItem"
    didl_properties_defs = Item.didl_properties_defs + [
        ("upnp", "bookmarkedObjectID", "R"),
        ("upnp", "neverPlayable", "O"),
        ("upnp", "deviceUDN", "R"),
        ("upnp", "serviceType", "R"),
        ("upnp", "serviceId", "R"),
        ("dc", "date", "O"),
        ("dc", "stateVariableCollection", "R"),
    ]


class EpgItem(Item):
    """DIDL EPG Item."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.epgItem"
    didl_properties_defs = Item.didl_properties_defs + [
        ("upnp", "channelGroupName", "O"),
        ("upnp", "channelGroupName@id", "O"),
        ("upnp", "epgProviderName", "O"),
        ("upnp", "serviceProvider", "O"),
        ("upnp", "channelName", "O"),
        ("upnp", "channelNr", "O"),
        ("upnp", "programTitle", "O"),
        ("upnp", "seriesTitle", "O"),
        ("upnp", "programID", "O"),
        ("upnp", "programID@type", "O"),
        ("upnp", "seriesID", "O"),
        ("upnp", "seriesID@type", "O"),
        ("upnp", "channelID", "O"),
        ("upnp", "channelID@type", "O"),
        ("upnp", "channelID@distriNetworkName", "O"),
        ("upnp", "channelID@distriNetworkID", "O"),
        ("upnp", "episodeType", "O"),
        ("upnp", "episodeCount", "O"),
        ("upnp", "episodeNumber", "O"),
        ("upnp", "episodeSeason", "O"),
        ("upnp", "programCode", "O"),
        ("upnp", "programCode@type", "O"),
        ("upnp", "rating", "O"),
        ("upnp", "rating@type", "O"),
        ("upnp", "rating@advice", "O"),
        ("upnp", "rating@equivalentAge", "O"),
        ("upnp", "recommendationID", "O"),
        ("upnp", "recommendationID@type", "O"),
        ("upnp", "genre", "O"),
        ("upnp", "genre@id", "O"),
        ("upnp", "genre@extended", "O"),
        ("upnp", "artist", "O"),
        ("upnp", "artist@role", "O"),
        ("upnp", "actor", "O"),
        ("upnp", "actor@role", "O"),
        ("upnp", "author", "O"),
        ("upnp", "author@role", "O"),
        ("upnp", "producer", "O"),
        ("upnp", "director", "O"),
        ("dc", "publisher", "O"),
        ("dc", "contributor", "O"),
        ("upnp", "callSign", "O"),
        ("upnp", "networkAffiliation", "O"),
        # ('upnp', 'serviceProvider', 'O'),  # duplicate in standard
        ("upnp", "price", "O"),
        ("upnp", "price@currency", "O"),
        ("upnp", "payPerView", "O"),
        # ('upnp', 'epgProviderName', 'O'),  # duplicate in standard
        ("dc", "description", "O"),
        ("upnp", "longDescription", "O"),
        ("upnp", "icon", "O"),
        ("upnp", "region", "O"),
        ("upnp", "rights", "O"),
        ("dc", "language", "O"),
        ("dc", "relation", "O"),
        ("upnp", "scheduledStartTime", "O"),
        ("upnp", "scheduledEndTime", "O"),
        ("upnp", "recordable", "O"),
        ("upnp", "foreignMetadata", "O"),
    ]


class AudioProgram(EpgItem):
    """DIDL Audio Program."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.epgItem.audioProgram"
    didl_properties_defs = Item.didl_properties_defs + [
        ("upnp", "radioCallSign", "O"),
        ("upnp", "radioStationID", "O"),
        ("upnp", "radioBand", "O"),
    ]


class VideoProgram(EpgItem):
    """DIDL Video Program."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.item.epgItem.videoProgram"
    didl_properties_defs = Item.didl_properties_defs + [
        ("upnp", "price", "O"),
        ("upnp", "price@currency", "O"),
        ("upnp", "payPerView", "O"),
    ]


# endregion


# region: containers
class Container(DidlObject, list):
    """DIDL Container."""

    # pylint: disable=too-few-public-methods

    tag = "container"
    upnp_class = "object.container"
    didl_properties_defs = DidlObject.didl_properties_defs + [
        ("didl_lite", "@childCount", "O"),
        ("upnp", "createClass", "O"),
        ("upnp", "searchClass", "O"),
        ("didl_lite", "@searchable", "O"),
        ("didl_lite", "@neverPlayable", "O"),
    ]

    def __init__(
        self,
        id: str = "",
        parent_id: str = "",
        descriptors: Optional[Sequence["Descriptor"]] = None,
        xml_el: Optional[ET.Element] = None,
        strict: bool = True,
        children: Iterable[DidlObject] = (),
        **properties: Any,
    ) -> None:
        """Initialize."""
        # pylint: disable=redefined-builtin,too-many-arguments,too-many-positional-arguments
        super().__init__(id, parent_id, descriptors, xml_el, strict, **properties)
        self.extend(children)

    @classmethod
    def from_xml(cls: Type[TC], xml_el: ET.Element, strict: bool = True) -> TC:
        """
        Initialize from an XML node.

        I.e., parse XML and return instance.
        """
        instance = super().from_xml(xml_el, strict)

        # add all children
        didl_objects = from_xml_el(xml_el, strict)
        instance.extend(didl_objects)  # pylint: disable=no-member

        return instance

    def to_xml(self) -> ET.Element:
        """Convert self to XML Element."""
        container_el = super().to_xml()

        for didl_object in self:
            didl_object_el = didl_object.to_xml()
            container_el.append(didl_object_el)

        return container_el

    def __repr__(self) -> str:
        """Evaluatable string representation of this object."""
        class_name = type(self).__name__
        attr = ", ".join(f"{key}={val!r}" for key, val in self.__dict__.items() if key not in ("class", "xml_el"))
        children_repr = ", ".join(repr(child) for child in self)
        return f"{class_name}({attr}, children=[{children_repr}])"


class Person(Container):
    """DIDL Person."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.person"
    didl_properties_defs = Container.didl_properties_defs + [
        ("dc", "language", "O"),
    ]


class MusicArtist(Person):
    """DIDL Music Artist."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.person.musicArtist"
    didl_properties_defs = Container.didl_properties_defs + [
        ("upnp", "genre", "O"),
        ("upnp", "artistDiscographyURI", "O"),
    ]


class PlaylistContainer(Container):
    """DIDL Playlist Container."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.playlistContainer"
    didl_properties_defs = Container.didl_properties_defs + [
        ("upnp", "artist", "O"),
        ("upnp", "genre", "O"),
        ("upnp", "longDescription", "O"),
        ("upnp", "producer", "O"),
        ("upnp", "storageMedium", "O"),
        ("dc", "description", "O"),
        ("dc", "contributor", "O"),
        ("dc", "date", "O"),
        ("dc", "language", "O"),
        ("dc", "rights", "O"),
    ]


class Album(Container):
    """DIDL Album."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.album"
    didl_properties_defs = Container.didl_properties_defs + [
        ("upnp", "storageMedium", "O"),
        ("dc", "longDescription", "O"),
        ("dc", "description", "O"),
        ("dc", "publisher", "O"),
        ("dc", "contributor", "O"),
        ("dc", "date", "O"),
        ("dc", "relation", "O"),
        ("dc", "rights", "O"),
    ]


class MusicAlbum(Album):
    """DIDL Music Album."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.album.musicAlbum"
    didl_properties_defs = Container.didl_properties_defs + [
        ("upnp", "artist", "O"),
        ("upnp", "genre", "O"),
        ("upnp", "producer", "O"),
        ("upnp", "albumArtURI", "O"),
        ("upnp", "toc", "O"),
    ]


class PhotoAlbum(Album):
    """DIDL Photo Album."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.album.photoAlbum"
    didl_properties_defs = Container.didl_properties_defs + []


class Genre(Container):
    """DIDL Genre."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.genre"
    didl_properties_defs = Container.didl_properties_defs + [
        ("upnp", "genre", "O"),
        ("upnp", "longDescription", "O"),
        ("dc", "description", "O"),
    ]


class MusicGenre(Genre):
    """DIDL Music Genre."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.genre.musicGenre"
    didl_properties_defs = Container.didl_properties_defs + []


class MovieGenre(Genre):
    """DIDL Movie Genre."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.genre.movieGenre"
    didl_properties_defs = Container.didl_properties_defs + []


class ChannelGroup(Container):
    """DIDL Channel Group."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.channelGroup"
    didl_properties_defs = Container.didl_properties_defs + [
        ("upnp", "channelGroupName", "O"),
        ("upnp", "channelGroupName@id", "O"),
        ("upnp", "epgProviderName", "O"),
        ("upnp", "serviceProvider", "O"),
        ("upnp", "icon", "O"),
        ("upnp", "region", "O"),
    ]


class AudioChannelGroup(ChannelGroup):
    """DIDL Audio Channel Group."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.channelGroup.audioChannelGroup"
    didl_properties_defs = Container.didl_properties_defs + []


class VideoChannelGroup(ChannelGroup):
    """DIDL Video Channel Group."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.channelGroup.videoChannelGroup"
    didl_properties_defs = Container.didl_properties_defs + []


class EpgContainer(Container):
    """DIDL EPG Container."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.epgContainer"
    didl_properties_defs = Container.didl_properties_defs + [
        ("upnp", "channelGroupName", "O"),
        ("upnp", "channelGroupName@id", "O"),
        ("upnp", "epgProviderName", "O"),
        ("upnp", "serviceProvider", "O"),
        ("upnp", "channelName", "O"),
        ("upnp", "channelNr", "O"),
        ("upnp", "channelID", "O"),
        ("upnp", "channelID@type", "O"),
        ("upnp", "radioCallSign", "O"),
        ("upnp", "radioStationID", "O"),
        ("upnp", "radioBand", "O"),
        ("upnp", "callSign", "O"),
        ("upnp", "networkAffiliation", "O"),
        # ('upnp', 'serviceProvider', 'O'),  # duplicate in standard
        ("upnp", "price", "O"),
        ("upnp", "price@currency", "O"),
        ("upnp", "payPerView", "O"),
        # ('upnp', 'epgProviderName', 'O'),  # duplicate in standard
        ("upnp", "icon", "O"),
        ("upnp", "region", "O"),
        ("dc", "language", "O"),
        ("dc", "relation", "O"),
        ("upnp", "dateTimeRange", "O"),
    ]


class StorageSystem(Container):
    """DIDL Storage System."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.storageSystem"
    didl_properties_defs = Container.didl_properties_defs + [
        ("upnp", "storageTotal", "R"),
        ("upnp", "storageUsed", "R"),
        ("upnp", "storageFree", "R"),
        ("upnp", "storageMaxPartition", "R"),
        ("upnp", "storageMedium", "R"),
    ]


class StorageVolume(Container):
    """DIDL Storage Volume."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.storageVolume"
    didl_properties_defs = Container.didl_properties_defs + [
        ("upnp", "storageTotal", "R"),
        ("upnp", "storageUsed", "R"),
        ("upnp", "storageFree", "R"),
        ("upnp", "storageMedium", "R"),
    ]


class StorageFolder(Container):
    """DIDL Storage Folder."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.storageFolder"
    didl_properties_defs = Container.didl_properties_defs + [
        ("upnp", "storageUsed", "R"),
    ]


class BookmarkFolder(Container):
    """DIDL Bookmark Folder."""

    # pylint: disable=too-few-public-methods

    upnp_class = "object.container.bookmarkFolder"
    didl_properties_defs = Container.didl_properties_defs + [
        ("upnp", "genre", "O"),
        ("upnp", "longDescription", "O"),
        ("dc", "description", "O"),
    ]


# endregion


class Resource:
    """DIDL Resource."""

    # pylint: disable=too-few-public-methods,too-many-instance-attributes

    def __init__(
        self,
        uri: Optional[str],
        protocol_info: Optional[str],
        import_uri: Optional[str] = None,
        size: Optional[str] = None,
        duration: Optional[str] = None,
        bitrate: Optional[str] = None,
        sample_frequency: Optional[str] = None,
        bits_per_sample: Optional[str] = None,
        nr_audio_channels: Optional[str] = None,
        resolution: Optional[str] = None,
        color_depth: Optional[str] = None,
        protection: Optional[str] = None,
        xml_el: Optional[ET.Element] = None,
    ) -> None:
        """Initialize."""
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        self.uri = uri
        self.protocol_info = protocol_info
        self.import_uri = import_uri
        self.size = size
        self.duration = duration
        self.bitrate = bitrate
        self.sample_frequency = sample_frequency
        self.bits_per_sample = bits_per_sample
        self.nr_audio_channels = nr_audio_channels
        self.resolution = resolution
        self.color_depth = color_depth
        self.protection = protection
        self.xml_el = xml_el

    @classmethod
    def from_xml(cls: Type[TR], xml_el: ET.Element) -> TR:
        """Initialize from an XML node."""
        uri = xml_el.text
        protocol_info = xml_el.attrib.get("protocolInfo")
        import_uri = xml_el.attrib.get("importUri")
        size = xml_el.attrib.get("size")
        duration = xml_el.attrib.get("duration")
        bitrate = xml_el.attrib.get("bitrate")
        sample_frequency = xml_el.attrib.get("sampleFrequency")
        bits_per_sample = xml_el.attrib.get("bitsPerSample")
        nr_audio_channels = xml_el.attrib.get("nrAudioChannels")
        resolution = xml_el.attrib.get("resolution")
        color_depth = xml_el.attrib.get("colorDepth")
        protection = xml_el.attrib.get("protection")
        return cls(
            uri,
            protocol_info=protocol_info,
            import_uri=import_uri,
            size=size,
            duration=duration,
            bitrate=bitrate,
            sample_frequency=sample_frequency,
            bits_per_sample=bits_per_sample,
            nr_audio_channels=nr_audio_channels,
            resolution=resolution,
            color_depth=color_depth,
            protection=protection,
            xml_el=xml_el,
        )

    def to_xml(self) -> ET.Element:
        """Convert self to XML."""
        attribs = {
            "protocolInfo": self.protocol_info or "",
        }
        res_el = ET.Element("res", attribs)
        res_el.text = self.uri
        return res_el

    def __repr__(self) -> str:
        """Evaluatable string representation of this object."""
        class_name = type(self).__name__
        attr = ", ".join(f"{key}={val!r}" for key, val in self.__dict__.items() if val is not None and key != "xml_el")
        return f"{class_name}({attr})"


class Descriptor:
    """DIDL Descriptor."""

    def __init__(
        self,
        id: str,
        name_space: str,
        type: Optional[str] = None,
        text: Optional[str] = None,
        xml_el: Optional[ET.Element] = None,
    ) -> None:
        """Initialize."""
        # pylint: disable=invalid-name,redefined-builtin,too-many-arguments,too-many-positional-arguments
        self.id = id
        self.name_space = name_space
        self.type = type
        self.text = text
        self.xml_el = xml_el

    @classmethod
    def from_xml(cls: Type[TD], xml_el: ET.Element) -> TD:
        """Initialize from an XML node."""
        id_ = xml_el.attrib["id"]
        name_space = xml_el.attrib["nameSpace"]
        type_ = xml_el.attrib.get("type")
        text = xml_el.text
        return cls(id_, name_space, type=type_, text=text, xml_el=xml_el)

    def to_xml(self) -> ET.Element:
        """Convert self to XML."""
        attribs = {
            "id": self.id,
            "nameSpace": self.name_space,
        }
        if self.type is not None:
            attribs["type"] = self.type
        desc_el = ET.Element("desc", attribs)
        desc_el.text = self.text
        return desc_el

    def __getattr__(self, name: str) -> Any:
        """Get attribute."""
        if name not in self.__dict__:
            raise AttributeError(name)
        return self.__dict__[name]

    def __repr__(self) -> str:
        """Evaluatable string representation of this object."""
        class_name = type(self).__name__
        attr = ", ".join(f"{key}={val!r}" for key, val in self.__dict__.items() if val is not None and key != "xml_el")
        return f"{class_name}({attr})"


# endregion


def to_xml_string(*objects: DidlObject) -> bytes:
    """Convert items to DIDL-Lite XML string."""
    root_el = ET.Element("DIDL-Lite", {})
    root_el.attrib["xmlns"] = NAMESPACES["didl_lite"]
    root_el.attrib["xmlns:dc"] = NAMESPACES["dc"]
    root_el.attrib["xmlns:upnp"] = NAMESPACES["upnp"]
    root_el.attrib["xmlns:sec"] = NAMESPACES["sec"]

    for didl_object in objects:
        didl_object_el = didl_object.to_xml()
        root_el.append(didl_object_el)

    return ET.tostring(root_el)


def from_xml_string(xml_string: str, strict: bool = True) -> List[Union[DidlObject, Descriptor]]:
    """Parse DIDL-Lite XML string."""
    if not strict:
        # Find all prefixes used in tags, e.g., <prefix:tag ...>
        used_prefixes = set(re.findall(r"<([A-Za-z_][\w.-]*):", xml_string))

        # Find all defined namespaces, e.g., xmlns:prefix=...
        defined_prefixes = set(re.findall(r"xmlns:([A-Za-z_][\w.-]*)=", xml_string))

        # Identify prefixes used but not defined.
        missing_prefixes = used_prefixes - defined_prefixes - {"DIDL-Lite", "dc", "upnp", "dlna"}

        if missing_prefixes:
            # Inject temporary namespace declarations into the DIDL-Lite root
            # opening tag. Anchoring on the root element (always present in
            # valid DIDL-Lite documents) instead of an optional `xmlns:dlna`
            # allows us to recover XML from devices that omit the dlna
            # namespace declaration entirely (observed on JBL Authentics and
            # WiiM/LinkPlay players sending `<song:*>` without `xmlns:dlna`).
            # The root itself may carry a prefix, as in `<didl:DIDL-Lite>`, so
            # the match is captured and written back unchanged.
            injections = " ".join(
                f'xmlns:{prefix}="http://tempuri.org/{prefix}/"' for prefix in sorted(missing_prefixes)
            )
            xml_string = re.sub(
                r"<((?:[A-Za-z_][\w.-]*:)?DIDL-Lite)\b",
                lambda match: f"<{match.group(1)} {injections}",
                xml_string,
                count=1,
            )

    # Proceed with parsing using the (potentially) patched xml_string
    xml_el = defusedxml.ElementTree.fromstring(xml_string)
    return from_xml_el(xml_el, strict)


def from_xml_el(xml_el: ET.Element, strict: bool = True) -> List[Union[DidlObject, Descriptor]]:
    """Convert XML Element to DIDL Objects."""
    didl_objects = []  # type: List[Union[DidlObject, Descriptor]]

    # items and containers, in order
    for child_el in xml_el:
        if child_el.tag != expand_namespace_tag("didl_lite:item") and child_el.tag != expand_namespace_tag(
            "didl_lite:container"
        ):
            continue

        # construct item
        upnp_class = child_el.find("./upnp:class", NAMESPACES)
        if upnp_class is None or not upnp_class.text:
            if strict:
                continue
            # WiiM Pro and possibly other Linkplay devices emit
            # upnp_class above the item element instead of inside it
            upnp_class = xml_el.find("./upnp:class", NAMESPACES)
            if upnp_class is None or not upnp_class.text:
                continue
        didl_object_type = type_by_upnp_class(upnp_class.text, strict)
        if didl_object_type is None:
            if strict:
                raise DidlLiteException(f"upnp:class {upnp_class.text} is unknown")
            continue
        didl_object = didl_object_type.from_xml(child_el, strict)
        didl_objects.append(didl_object)

    # descriptors
    for desc_el in xml_el.findall("./didl_lite:desc", NAMESPACES):
        desc = Descriptor.from_xml(desc_el)
        didl_objects.append(desc)

    return didl_objects


# upnp_class to python type mapping
def type_by_upnp_class(upnp_class: str, strict: bool = True) -> Optional[Type[DidlObject]]:
    """Get DidlObject-type by upnp_class.

    When strict is False, the upnp_class lookup will be done ignoring string
    case.
    """
    if strict:
        return _upnp_class_map.get(upnp_class)
    return _upnp_class_map_lowercase.get(upnp_class.lower())
